import tweepy
import pandas as pd
from typing import List
import datetime
import pydash
from numpy import inf

tweet_fields = ["author_id,created_at,source,public_metrics"]
tweet_personal_fields = [
    "author_id,created_at,source,public_metrics,non_public_metrics,organic_metrics"
]


class Twitter:

    # Authenticate as an app.
    __bearer_token: str

    # Authenticate as a user.
    __consumer_key: str
    __consumer_secret: str
    __access_token: str
    __access_token_secret: str

    # Twitter v2 auth
    __app_client: tweepy.Client
    __user_client: tweepy.Client

    __me: pd.Series

    def connect(
        self,
        bearer_token: str,
        consumer_key: str,
        consumer_secret: str,
        access_token: str,
        access_token_secret: str,
    ) -> "Twitter":
        self.__bearer_token = bearer_token

        self.__app_client = tweepy.Client(bearer_token=self.__bearer_token)

        self.__consumer_key = consumer_key
        self.__consumer_secret = consumer_secret
        self.__access_token = access_token
        self.__access_token_secret = access_token_secret

        self.__user_client = tweepy.Client(
            consumer_key=consumer_key,
            consumer_secret=consumer_secret,
            access_token=access_token,
            access_token_secret=access_token_secret,
        )

        self.__me = self.get_me()

        return self

    @property
    def app_client(self):
        return self.__app_client

    @property
    def user_client(self):
        return self.__user_client

    def get_user(self, username: str) -> pd.Series:
        users = self.__app_client.get_users(usernames=[username])
        if users is None:
            return None
        return pd.Series(users.data[0].data)

    def get_me(self):
        me = self.__user_client.get_me()
        if me is None:
            return None
        return pd.Series(me.data.data)

    def get_my_tweets(self, **kwargs):
        return self.get_users_tweets(self.__me.id, **kwargs)

    def get_users_tweets(
        self,
        user_id: str,
        tweet_count=200,
        tweet_fields: List[str] = tweet_fields,
        start_time=datetime.datetime.now() - datetime.timedelta(days=30),
        end_time=datetime.datetime.now(),
    ) -> pd.DataFrame:
        should_stop = False
        tweets_array = []
        next_token = None

        while len(tweets_array) < tweet_count and should_stop is False:
            tweets_left_to_fetch = tweet_count - len(tweets_array)

            if tweets_left_to_fetch > 100:
                max_results = 100
            elif tweets_left_to_fetch < 5:
                max_results = 5
            else:
                max_results = tweets_left_to_fetch
            tweets = self.__app_client.get_users_tweets(
                id=user_id,
                max_results=max_results,
                start_time=start_time,
                end_time=end_time,
                pagination_token=next_token,
            )
            next_token = pydash.get(tweets, "meta.next_token", None)
            if next_token is None:
                should_stop = True
            is_own_tweets = user_id == self.__me.id
            for tweet in tweets.data:
                tweet_id = tweet.id

                if is_own_tweets is True:
                    rich_tweet_response = self.__user_client.get_tweet(
                        tweet_id, tweet_fields=tweet_personal_fields, user_auth=True
                    )
                    if len(rich_tweet_response.errors):
                        rich_tweet_response = self.__user_client.get_tweet(
                            tweet_id, tweet_fields=tweet_fields, user_auth=True
                        )
                else:
                    rich_tweet_response = self.__app_client.get_tweet(
                        tweet_id, tweet_fields=tweet_fields, user_auth=False
                    )
                rtd = rich_tweet_response.data

                tweets_array.append(
                    {
                        "TWEET_ID": rtd.id,
                        "TWEET_URL": f"https://twitter.com/{self.__me.username}/status/{rtd.id}",
                        "CREATED_AT": rtd.created_at,
                        "AUTHOR_ID": rtd.author_id,
                        "AUTHOR_NAME": self.__me["name"],
                        "AUTHOR_USERNAME": self.__me.username,
                        "TEXT": rtd.text,
                        "PUBLIC_RETWEETS": pydash.get(
                            rtd, "public_metrics.retweet_count", 0
                        ),
                        "PUBLIC_REPLIES": pydash.get(
                            rtd, "public_metrics.reply_count", 0
                        ),
                        "PUBLIC_LIKES": pydash.get(rtd, "public_metrics.like_count", 0),
                        "PUBLIC_QUOTES": pydash.get(
                            rtd, "public_metrics.quote_count", 0
                        ),
                        "ORGANIC_RETWEETS": pydash.get(
                            rtd, "organic_metrics.retweet_count", 0
                        ),
                        "ORGANIC_REPLIES": pydash.get(
                            rtd, "organic_metrics.reply_count", 0
                        ),
                        "ORGANIC_LIKES": pydash.get(
                            rtd, "organic_metrics.like_count", 0
                        ),
                        "ORGANIC_QUOTES": pydash.get(
                            rtd, "organic_metrics.quote_count", 0
                        ),
                        "USER_PROFILE_CLICKS": pydash.get(
                            rtd, "non_public_metrics.user_profile_clicks", 0
                        ),
                        "IMPRESSIONS": pydash.get(
                            rtd, "non_public_metrics.impression_count", 0
                        ),
                    }
                )
        # Create final dataframe
        as_types = {
            "PUBLIC_RETWEETS": int,
            "PUBLIC_REPLIES": int,
            "PUBLIC_LIKES": int,
            "PUBLIC_QUOTES": int,
            "ORGANIC_RETWEETS": int,
            "ORGANIC_REPLIES": int,
            "ORGANIC_LIKES": int,
            "ORGANIC_QUOTES": int,
            "USER_PROFILE_CLICKS": int,
            "IMPRESSIONS": int,
        }
        df = pd.DataFrame(tweets_array).astype(as_types)
        df["ENGAGEMENTS"] = (
            df["PUBLIC_RETWEETS"]
            + df["PUBLIC_REPLIES"]
            + df["PUBLIC_LIKES"]
            + df["PUBLIC_QUOTES"]
            + df["USER_PROFILE_CLICKS"]
        )
        df["ENGAGEMENT_RATE"] = df["ENGAGEMENTS"] / df["IMPRESSIONS"]
        df = df.round({"ENGAGEMENT_RATE": 4})
        df = df.fillna({"ENGAGEMENT_RATE": 0})
        df["ENGAGEMENT_RATE"] = df["ENGAGEMENT_RATE"].replace(inf, 0)
        df["ENGAGEMENT_RATE"] = df["ENGAGEMENT_RATE"].apply(lambda x: 0 if x < 0 else x)
        return df.reset_index(drop=True)



import pandas as pd
import requests
from datetime import datetime

DATE_FORMAT = "%Y-%m-%d"
DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"
NUMBER_FORMAT = "{:,.2f} €"


class BudgetInsight:
    def connect(self, domain, client_id, client_secret, uuid=None):
        # Init atributes
        self.api_url = f"https://{domain}/2.0"

        if uuid is not None:
            req_url = f"{self.api_url}/auth/renew"
            res = requests.post(
                req_url,
                data={
                    "client_id": client_id,
                    "client_secret": client_secret,
                    "id_user": uuid,
                },
            )
            try:
                res.raise_for_status()
            except requests.HTTPError as e:
                return e
            res_json = res.json()
            token = res_json.get("access_token")
        else:
            req_url = f"{self.api_url}/auth/init"
            res = requests.post(
                req_url, data={"client_id": client_id, "client_secret": client_secret}
            )
            try:
                res.raise_for_status()
            except requests.HTTPError as e:
                return e
            res_json = res.json()
            id_user = res_json.get("id_user")
            print(f"User created. Keep your user id for next connection : {id_user}")
            token = res_json.get("auth_token")
        # Init headers
        self.headers = {"authorization": f"Bearer {token}"}

        # Init end point
        self.connections = Connections(self.api_url, self.headers)
        self.accounts = Accounts(self.api_url, self.headers)
        self.transactions = Transactions(self.api_url, self.headers)
        self.connectors = Connectors(self.api_url)

        # Set connexion to active
        self.connected = True
        return self


class Connectors(BudgetInsight):
    def __init__(self, api_url):
        BudgetInsight.__init__(self)
        self.api_url = api_url

    def get(self):
        req_url = f"{self.api_url}/connectors/"
        res = requests.get(req_url)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            return e
        res_json = res.json()
        connectors = res_json.get("connectors")
        df = pd.DataFrame(connectors)
        return df

    def get_fields(self, bq_id):
        req_url = f"{self.api_url}/connectors/{bq_id}/fields"
        res = requests.get(req_url)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            return e
        res_json = res.json()
        fields = res_json.get("fields")
        return fields

    def get_referential(self):
        df = Connectors.get(self)
        df = df[["id", "name"]]
        df = df.rename(columns={"id": "id_bank", "name": "bank_name"})
        return df


class Connections(BudgetInsight):
    def __init__(self, api_url, headers):
        BudgetInsight.__init__(self)
        self.api_url = api_url
        self.headers = headers

    def create(self, data):
        req_url = f"{self.api_url}/users/me/connections"
        res = requests.post(req_url, data=data, headers=self.headers)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            return e
        res_json = res.json()
        return res_json

    def update(self, data, connection_id):
        req_url = f"{self.api_url}/users/me/connections/{connection_id}"
        res = requests.post(req_url, data=data, headers=self.headers)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            return e
        res_json = res.json()
        return res_json

    def force(self, connection_id):
        req_url = f"{self.api_url}/users/me/connections/{connection_id}"
        res = requests.put(req_url, headers=self.headers)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            return e
        res_json = res.json()
        return res_json

    def get(self):
        req_url = f"{self.api_url}/users/me/connections"
        res = requests.get(req_url, headers=self.headers)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            return e
        res_json = res.json()
        connections = res_json.get("connections")
        df = pd.DataFrame(connections)
        return df

    def delete(self, connection_id):
        req_url = f"{self.api_url}/users/me/connections/{connection_id}"
        res = requests.delete(req_url, headers=self.headers)
        try:
            res.raise_for_status()
        except ValueError:
            print(
                f"❌ Connection (id={connection_id}) does not exist. Please enter a valid id."
            )
            return self.get()
        return f"✔️ Connection (id={connection_id}) successfully deleted."

    def get_referential(self):
        # Get ref connector
        ref_connectors = Connectors.get_referential(self)

        # Get connections
        df = Connections.get(self)
        df = df[["id", "id_user", "id_bank", "state", "error"]]

        # Create ref connections
        df = pd.merge(df, ref_connectors, on="id_bank")
        df = df.rename(
            columns={
                "id": "id_connection",
                "state": "connection_state",
                "error": "connection_error",
            }
        )
        return df


class Accounts(BudgetInsight):
    def __init__(self, api_url, headers):
        BudgetInsight.__init__(self)
        self.api_url = api_url
        self.headers = headers

    def get(self, account_id=None):
        if account_id is None:
            req_url = f"{self.api_url}/users/me/accounts?all"
        else:
            req_url = f"{self.api_url}/users/me/accounts/{account_id}?all"
        res = requests.get(req_url, headers=self.headers)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            return e
        res_json = res.json()
        accounts = res_json.get("accounts")
        if accounts is None:
            return res_json
        else:
            return pd.DataFrame(accounts)

    def update(self, account_id, data=None):
        req_url = f"{self.api_url}/users/me/accounts/{account_id}?all"
        res = requests.put(req_url, data=data, headers=self.headers)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            return e
        res_json = res.json()
        return res_json

    def enable(self, account_id):
        account = self.get(account_id)
        if account.get("disabled") is not None:
            data = {"disabled": False}
            self.update(account_id, data)
            return f"✔️ Account (id={account_id}) enabled. You can now access to all your transactions."
        else:
            return "Account already enabled."

    def enable_all(self):
        enables = []
        df = self.get()
        for _, row in df.iterrows():
            account_id = row.id
            number = row.number
            name = row.original_name
            disabled = row.disabled
            if disabled is None:
                self.enable(account_id)
                enable = "Account already enabled."
            else:
                enable = "✔️ Account enabled"
            data = {
                "ACCOUNT_ID": account_id,
                "ACCOUNT_NUMBER": number,
                "ACCOUNT_NAME": name,
                "ENABLE": enable,
            }
            enables.append(data)
        return pd.DataFrame(enables)

    def get_referential(self):
        # Get ref connections
        ref_connections = Connections.get_referential(self)

        # Get accounts
        df_accounts = Accounts.get(self)

        # Create ref accounts
        ref_accounts = pd.merge(
            df_accounts, ref_connections, on=["id_connection", "id_user"]
        )
        to_keep = [
            "id_user",
            "id_connection",
            "connection_state",
            "connection_error",
            "id",
            "number",
            "original_name",
            "type",
            "iban",
            "bic",
            "id_bank",
            "bank_name",
        ]
        ref_accounts = ref_accounts[to_keep]
        to_rename = {
            "id": "id_account",
            "number": "account_number",
            "original_name": "account_name",
            "type": "account_type",
        }
        ref_accounts = ref_accounts.rename(columns=to_rename)
        ref_accounts = ref_accounts.drop_duplicates(
            subset=["account_number"]
        ).reset_index(drop=True)
        return ref_accounts

    def get_statement(self):
        # Get accounts
        df = Accounts.get(self)
        to_rename = {"id": "id_account"}
        df = df.rename(columns=to_rename)

        # Get ref accounts
        ref = Accounts.get_referential(self)

        # Create df positions
        df_statement = pd.merge(df, ref, on=["id_connection", "id_user", "id_account"])

        to_keep = [
            "id_connection",
            "connection_state",
            "bank_name",
            "id_account",
            "account_number",
            "account_name",
            "account_type",
            "balance",
            "last_update",
        ]
        df_statement = df_statement[to_keep]
        df_statement.columns = df_statement.columns.str.upper()
        df_statement = df_statement.drop_duplicates(
            subset=["ACCOUNT_NUMBER"]
        ).reset_index(drop=True)
        df_statement["DATE_EXTRACTION"] = datetime.now().strftime(DATETIME_FORMAT)
        return df_statement


class Transactions(BudgetInsight):
    def __init__(self, api_url, headers):
        BudgetInsight.__init__(self)
        self.api_url = api_url
        self.headers = headers

    def get(self):
        req_url = f"{self.api_url}/users/me/transactions?all"
        res = requests.get(req_url, headers=self.headers)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            return e
        res_json = res.json()
        transactions = res_json.get("transactions")
        df = pd.DataFrame(transactions)
        return df

    def get_statement(self):
        # Get transactions
        df = Transactions.get(self)
        to_rename = {
            "id": "id_transaction",
            "original_wording": "transaction_label",
            "type": "transaction_type",
            "value": "amount",
        }
        df = df.rename(columns=to_rename)

        # Get ref accounts
        ref = Accounts.get_referential(self)

        # Create statement
        df_statement = pd.merge(df, ref, on="id_account")

        # Select column to keep
        to_keep = [
            "account_number",
            "account_name",
            "bank_name",
            "id_transaction",
            "date",
            "transaction_label",
            "transaction_type",
            "amount",
        ]
        df_statement = df_statement[to_keep]
        df_statement.columns = df_statement.columns.str.upper()
        df_statement = df_statement.drop_duplicates(
            subset=["ACCOUNT_NUMBER", "ID_TRANSACTION"]
        ).reset_index(drop=True)
        df_statement["DATE_EXTRACTION"] = datetime.now().strftime(DATETIME_FORMAT)
        return df_statement



from naas_drivers.driver import InDriver
from IPython.core.display import display, HTML
import markdown2


class Markdown(InDriver):
    """Markdown generator lib"""

    def convert(self, data, output_type="html"):
        data_read = self.open_or_read(data)
        if output_type == "html":
            return markdown2.markdown(data_read)
        else:
            error_text = f"output {output_type} not supported for now"
            self.print_error(error_text)

    def display(self, data):
        data_read = self.open_or_read(data)
        display(HTML(self.convert(data_read)))



from os import path
from shareplum import Site
from shareplum import Office365
from shareplum.site import Version


class Sharepoint:
    def connect(self, endpoint: str, username: str, password: str, site: str):
        self.endpoint = endpoint

        self.authcookie = Office365(
            f"https://{endpoint}", username=username, password=password
        ).GetCookies()
        self.site = Site(
            f"https://{endpoint}/sites/{site}",
            version=Version.v2016,
            authcookie=self.authcookie,
        )

        return self

    def list_folder(self, folder: str):
        f = self.site.Folder(folder)
        return {"folders": f.folders, "files": f.files}

    def get_file(self, src: str, dst: str):
        try:
            folder = self.site.Folder(path.dirname(src))
            with open(dst, "wb") as f:
                f.write(folder.get_file(path.basename(src)))
                f.close()
            print(f'✅ "{src}" downloaded in "{dst}"')
        except Exception as e:
            print(f'[red]❌ Error while downloading "{src}" into "{dst}"')
            print(e)

    def upload_file(self, src: str, dst: str):
        try:
            folder = self.site.Folder(path.dirname(dst))
            with open(src, mode="rb") as file:
                fileContent = file.read()
                folder.upload_file(fileContent, path.basename(dst))
            print(f'✅ "{src}" uploaded in "{dst}"')
        except Exception as e:
            print(f'[red]❌ Error while uploading "{src}" into "{dst}"')
            print(e)



from google.cloud import bigquery
from google.oauth2 import service_account
import pandas as pd
import logging


class BigQuery:
    @property
    def _logger(self):
        return logging.getLogger(__name__)

    def connect(self, service_account_file, project_id):
        try:
            credentials = service_account.Credentials.from_service_account_file(
                service_account_file
            )
            self.connection = bigquery.Client(
                credentials=credentials, project=project_id
            )
            self._logger.info("Connection established!")
            return self
        except Exception as e:
            self._logger.error(e)

    def execute_query(self, query):
        try:
            df = self.connection.query(query).to_dataframe()

            if not df.empty:
                return df
            else:
                return True
        except Exception as e:
            self._logger.error(e)
            return False

    def load_data_from_csv(self, dataset, table, csv_file):
        df = pd.read_csv(csv_file)

        dataset_ref = self.connection.dataset(dataset)
        table_ref = dataset_ref.table(table)

        self.connection.load_table_from_dataframe(df, table_ref).result()



from naas_drivers.driver import InDriver
import pandas as pd
from pandas import json_normalize
from typing import Union
from vaderSentiment.vaderSentiment import SentimentIntensityAnalyzer


class Sentiment(InDriver):
    def __sanitize_dataset(
        self, dataset: Union[pd.DataFrame, pd.Series], column_name: str
    ) -> pd.DataFrame:
        if isinstance(dataset, pd.Series):
            df = dataset.to_frame().reset_index()
        elif isinstance(dataset, pd.DataFrame):
            df = dataset
        else:
            error_text = f"The dataset should either be a Dataframe or a Series, not {dataset.__name__}"
            self.print_error(error_text)
            return None
        if column_name not in df.columns:
            error_text = f"The {column_name} is not in dataset columns."
            self.print_error(error_text)
            return None
        return df

    def categorize(self, comp_score):
        if float(comp_score) >= 0.05:
            return "Positive"
        elif float(comp_score) <= -0.05:
            return "Negative"
        else:
            return "Neutral"

    def __calculate_sentiment(
        self, df: pd.DataFrame, column_name: str, details: bool
    ) -> pd.DataFrame:
        analyzer = SentimentIntensityAnalyzer()
        df["senti_score"] = df[column_name].map(analyzer.polarity_scores)
        senti_df = json_normalize(df["senti_score"])

        senti_df["Sentiment"] = senti_df["compound"].map(self.categorize)
        senti_df = senti_df.rename(
            columns={
                "compound": "Score",
                "pos": "Positive",
                "neg": "Negative",
                "neu": "Neutral",
            }
        )
        df = pd.concat([df, senti_df], axis=1)
        cols = [column_name, "Sentiment", "Score"]
        if details:
            cols.extend(["Positive", "Negative", "Neutral"])
        df = df[cols]

        return df

    def get(
        self,
        dataset: Union[pd.DataFrame, pd.Series],
        column_name: str,
        details: bool = False,
    ) -> pd.DataFrame:
        """
        Calculates the sentiment of the text in each row in the dataset &
        returns it along with the input as a single dataframe in the output.
        :param dataset: either a pandas Dataframe or a Series
        :param column_name: String
                            The exact column name on which the sentiment is to be calculated .
                            This should be present in the dataset.
        :param details: Boolean, Default: False
                         Tells whether the details of sentiment analysis are to be outputted
        :return: pandas dataframe

        """
        df = self.__sanitize_dataset(dataset, column_name)
        df = self.__calculate_sentiment(df, column_name, details)

        return df
